module addflags_mod

use dal
use sla
use sasdal
use caloaldefs
use cal
use omconstants_mod
use omsourceparam_mod
use omsourceflags_mod
use omallocation_mod
use omellipse_mod
use omutility_mod


!****************************************************************************
! QFLAG source quality flags convention
! BIT NO             FLAGGING                         Integer
!
!   0                BAD PIXEL                                 1
!   1                READ-OUT STREAK                           2
!   2                SMOKE-RING                                4
!   3                SOURCE ON STAR-SPIKE                      8
!   4                Mod-8 source                             16
!   5                Source within central enhancement        32
!   6                Source lies near to a bright source      64
!   7                Near an edge                            128
!   8                point-source within extended source     256
!   9                Weird source                            512

!*****************************************************************************

!****************************************************************************
! CFLAG source confusion flags convention
! BIT NO             FLAGGING                                                Integer
!
!   0                one or more source within 0.5-1.0 aperture radius       1
!   1                one or more source within 0.5 aperture radius           2
!
!*****************************************************************************



!integer(int32), parameter, private               :: int32NullValue = -1      
!integer(int8), parameter, private                :: int8NullValue = 127      
!integer(int16), parameter, private               :: int16NullValue = 32767      


real(single), private, pointer :: myImage(:, :) => null()
integer(int16), private, pointer :: myQualityImage(:, :) => null()    


public

contains

!*********************************************************************
!
!
!
!*********************************************************************
subroutine createQualityMap(nx, ny)
    integer(int16), intent(in) :: nx, ny

    myqualityImage => allocate2DInt16(0_int16, nx - 1_int16, &
                     0_int16, ny - 1_int16, "qualityimage")

end subroutine createQualityMap
!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine createImage(nx, ny)
    integer(int16), intent(in) :: nx, ny

    myimage => allocate2DSingle(0_int16, nx - 1_int16, &
              0_int16, ny - 1_int16, "myimage")

end subroutine createImage

!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine setQualityPixel(i, j, value)
    integer(int16), intent(in) :: i, j, value

    myqualityImage(i, j) = value       

end subroutine setQualityPixel

!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine setImagePixel(i, j, value)
    integer(int16), intent(in) :: i, j

    real(single), intent(in) :: value

    myimage(i, j) = value       

end subroutine setImagePixel



!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine deleteImage
   
    deallocate(myimage)
    nullify(myimage)

end subroutine deleteImage


!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine deleteQualityMap
   
    deallocate(myqualityImage)
    nullify(myqualityImage)


end subroutine deleteQualityMap






!*********************************************************************
!
!
!
!
!*********************************************************************
integer(int16) Function getQualityPixel(i, j)
    integer(int16), intent(in) :: i, j

    getQualityPixel = myqualityImage(i, j)


end function getQualityPixel




!*********************************************************************
!
!
!
!
!*********************************************************************
real(single) Function getImagePixel(i, j)
    integer(int16), intent(in) :: i, j

    getImagePixel = myimage(i, j)


end function getImagePixel





!*********************************************************************
!
!
!
!
!*********************************************************************
subroutine createoutputlist



    call copyDataSet( trim( stringParameter("set") ), trim( stringParameter("outset") ))


end subroutine createoutputlist

!*********************************************************************
!
! Sets pixels on quality image
!
!
!
!*********************************************************************
subroutine addSourceFlags(nSources, source1, x0, y0, binX1)
    integer, intent(in) :: nsources
    type(sourceParameters) :: source1(1 : nSources)    
    integer, intent(in) :: x0, y0, binX1
    integer i 
    
    type(sourceParameters), pointer :: source(:) => null()
    real(single), pointer, dimension(:) :: xf=>null(), yf=>null()    
    type(dataSetT) :: dataSetH
    integer(int16) :: nx, ny
    integer(int8) :: spikeHalfWidth, halfWidth


    dataSetH = DataSet(trim(stringParameter("set")), read)

    call getImage(dataSetH, nx, ny, myImage)


    call release(dataSetH) 
    binX = binX1 
  

    allocate(source(1 : nsources))

    xf => allocateSingle(1, nSources, "XF")
    yf => allocateSingle(1, nSources, "YF")

    source(1 : nsources) = source1(1 : nsources)    


    !*************************************************************
    !
    ! Need the full frame coordinates
    !
    !************************************************************

    xf(1 : nSources) = source(1 : nSources)%xpos
    yf(1 : nSources) = source(1 : nSources)%ypos


    call ConvertCoordsToFullFrame(xF, yF, x0, y0, binx, binx)



    !*********************************************
    !  BIT 1 
    !
    !*********************************************
    call ReadOutStreaks3(nSources, source, myQualityImage)


    !*********************************************
    ! 
    ! BIT 2
    !
    !
    !*********************************************
    call SetSmokeRingsFlags(nSources,          &
                      source,                  &
                      xF, yF,                  &
                      myqualityImage, x0, y0, binX)
  


    !****************************************************************
    !
    ! Diffraction spikes
    ! BIT 3
    !
    !****************************************************************
    spikeHalfWidth = 8 / binX
    call StarSpikes1( nSources, source, xF, yF, myQualityImage, spikeHalfWidth)

 
    deallocate(xf, yf)
    nullify(xf, yf)


    !***********************************************************************
    !
    ! Set bit 4 (mod-8 pattern)
    !
    !***********************************************************************
    call SetMod8Flags( nSources, source, myImage, myQualityImage)


    !****************************************************************
    !
    ! Sources near much brighter ones
    ! BIT 6
    !
    !****************************************************************
    halfWidth = 30 / binX



!    write(*, *) "calling brightsourceproximityflags3"
!    call BrightSourceProximityFlags33(nSources, source, myQualityImage, halfWidth)


    !*****************************************************************
    !
    !
    !
    !*****************************************************************
!    call SetBit8Flags( nSources, source, myqualityImage)
  
 

    !*****************************************************************
    !
    !
    !*****************************************************************
    halfWidth = 12 / binX

    call SetBit9Flags( myimage, myqualityImage, halfWidth)




    deallocate(myImage, source)
    nullify(myImage, source)



end subroutine addSourceFlags






!***************************************************
!
! Subroutine SmokeRings
!
!
!***************************************************
Subroutine SetSmokeRingsFlags(n,               &
                      source,                  &
                      xF, yF, qualityMap, x0, y0, binX)
    integer, intent(in) :: n

    type(sourceParameters) :: source(:)    
    real(single) :: xF(:), yF(:)
    integer(int16), optional, pointer :: qualityMap(:, :)
    integer, intent(in) :: x0, y0, binx

    integer :: i, j
    real(single) :: xSmokeRing, ySmokeRing
    integer, parameter :: maxNoSmokeRings = 100
    integer :: noSmokeRings
    integer(int16) :: i0, j0, nx, ny

    real(single), parameter :: halfWidth = 38.0 ! IE 26 ring radius + 12 photometry radus)
  
 
  
    !*********************************************************************
    !
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 60.0 ! Was 70.0    


    if (n < 1) return

      
    if(present(qualityMap)) then
        nx = ubound(qualityMap, 1)
        ny = ubound(qualityMap, 2)


    end if

    noSmokeRings = 0

    !***********************************************************************
    !
    ! Loop through the sources to find any that may produce a smoke-ring
    !
    !
    !**********************************************************************

    do i = 1, n  
        if(source(i)%extensionFlag == 1) cycle
        if (source(i)%rawCountRate .gt. 0.0 .and. source(i)%rawCountRate .lt. 10000.0) then
            if ( source(i)%rawCountRate .ge. minimumCountRate .or. btest(source(i)%qualityFlag, 4) ) then ! This star may have produced a smoke ring




!                if ( btest(source(i)%qualityFlag, 2) ) then ! Then has been detected as a smoke ring 
!                           source(i)%qualityFlag = ibclr(source(i)%qualityFlag, 2)   ! BIT 2 UNSET
               
!                    cycle
!                end if


                !***********************************************************  
                ! The source is predicted to produce a smoke-ring 
                ! Predict where smoke-ring is
                ! Radial distance of star from centre of full-frame image
                ! xCentre and yCentre are defined in omconstants_mod (1024.5. 1024.5)
                !**************************************************************
               
               
                xsmokeRing = smokeRingXPosition(xF(i), yF(i))

                ysmokeRing = smokeRingYPosition(xF(i), yF(i))


                !***********************************************************************
                ! Need to convert back
                !
                !*********************************************************************** 
                call ConvertCoordsFromFullFrame(xSmokeRing, ySmokeRing, x0, y0, binx, binx)
           
                i0 = nint(xSmokeRing - 1.0)                
                j0 = nint(ySmokeRing - 1.0) 
                                 
                call setQualityMap(qualityMap, nx, ny, i0, j0, halfWidth / binX, 2)
            
                noSmokeRings = noSmokeRings + 1
              
                do j = 1, n
                    if (i .eq. j) cycle
                         
                    if ( RadialDistance(dble(xF(j)), dble(yF(j)), dble(xSmokeRing), dble(ySmokeRing)) .le. halfwidth) then
                        source(j)%qualityFlag = ibset(source(j)%qualityFlag, 2)   ! BIT 2 SET
               
                    end if

                end do
            end if
        end if   
    end do 

    write(errstr, '(a, i6)') "Number of smoke rings=", noSmokeRings
    call message(AppMsg, Verbose)
  
  
end subroutine setSmokeRingsFlags
!***********************************************************************
!
!
!
!***********************************************************************
subroutine setQualityMap(qualityMap, nx, ny, i0, j0, width, bitNumber)
    integer(int16), pointer :: qualityMap(:, :)
    integer(int16), intent(in) :: nx, ny, i0, j0
    real(single), intent(in) :: width
    integer, intent(in) :: bitNumber
    integer(int16) :: i, j, il, iu, jl, ju
    real(single) :: dx, dy, radius

    

 
    il = max(0_int16, i0 - int(width, int16) - 1_int16)
    iu = min(nx, i0 + int(width, int16) + 1_int16)
    jl = max(0_int16, j0 - int(width, int16) - 1_int16)
    ju = min(ny, j0 + int(width, int16) + 1_int16)


    do i = il, iu
        if(i < 0 .or. i > nx) cycle
        dx = abs(i0 - i)
        if(dx > width) cycle
        do j = jl, ju
            if(j < 0 .or. j > ny) cycle

            dy = abs(j0 - j)
            if(dy > width) cycle
            radius = sqrt(dx * dx + dy * dy)
            if(radius .le. width) then
                qualityMap(i, j) = ibset(qualityMap(i, j), bitNumber) 
            end if  

        end do
    end do


end subroutine setQualityMap

!***************************************************
!
!  subroutine BrightSourceProximityFlags
!
!***************************************************
Subroutine  BrightSourceProximityFlags33(nSources,      &
                      source, qualityMap, halfWidth)
  
    integer, intent(in) :: nSources
    type(sourceParameters), pointer :: source(:)

    integer(int16), optional, pointer :: qualityMap(:, :)
    integer(int8), intent(in), optional :: halfWidth 
    integer(int16) :: i, j, nx, ny, i0, j0
    
    real(single) :: radius, dx, dy, maxrad, countsRatio, rad1
  
    !*********************************************************************
    !
    ! The minimum count-rate for the setting of smoke-ring
    ! and diffraction spike flags
    !
    !*********************************************************************
    real(single), parameter :: minimumCountRate = 60.0    


    if(present(qualityMap)) then
        nx = ubound(qualityMap, 1)
        ny = ubound(qualityMap, 2)
    end if

    rad1 = 70.0



    do i = 1, nSources

        if ( btest(source(i)%qualityFlag, 4)   &          ! Bright-source
                  .or. source(i)%rawCountRate .ge. minimumCountRate ) then
 




            maxRad = min(rad1, 35.0 * source(i)%rawCountRate / minimumCountRate) / real(binx)
             
            !****************************************************************
            !
            ! Set the quality flags
            !
            !****************************************************************
            if(present(qualityMap) ) then


                i0 = max(0, nint( source(i)%xpos - 1.0))
                j0 = min(int(nx), nint( source(i)%ypos - 1.0) )

          
                call setQualityMap(qualityMap, nx, ny, i0, j0, maxRad, 6)

                
            end if


            do j = i - 1, 1, -1
                if ( btest(source(j)%qualityFlag, 4) )  cycle
                if ( btest(source(j)%qualityFlag, 6) )  cycle

  
                countsRatio = source(j)%rawCountRate / max(1.0, source(i)%rawCountRate)
                if (countsRatio > 1.0) cycle 



                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then

                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 6)

 

                end if
                  
            end do 

            do j = i + 1, nSources
                if ( btest(source(j)%qualityFlag, 4) )  cycle
                if ( btest(source(j)%qualityFlag, 6) )  cycle
  
                countsRatio = source(j)%rawCountRate / max(1.0, source(i)%rawCountRate)
                if (countsRatio > 1.0) cycle 

                dx = source(i)%xpos - source(j)%xpos
                dy = source(i)%ypos - source(j)%ypos
                   
                radius = sqrt(dx * dx + dy * dy)
                if (radius .le. maxRad) then
                    source(j)%qualityFlag = ibset(source(j)%qualityFlag, 6)
    
            
                    !****************************************************************
                    !
                    ! Set the quality flags
                    !
                    !****************************************************************
                    if(present(qualityMap) ) then


                        i0 = max(0, nint( source(j)%xpos - 1.0))
                        j0 = min(int(nx), nint( source(j)%ypos - 1.0) )
          
                        call setQualityMap(qualityMap, nx, ny, i0, j0, real(halfwidth, int8), 6)
                    end if 

                end if

            end do

        end if
    end do
    
end subroutine BrightSourceProximityFlags33 










!***************************************************
!
!  subroutine 
!
!***************************************************
Subroutine SetMod8Flags( nSources, source, image, qualityImage)
  
    integer, intent(in) :: nSources
    type(sourceParameters), pointer :: source(:)
    real(single), pointer :: image(:, :)
    integer(int16), pointer :: qualityImage(:, :)

    integer :: i
    if( .not. associated( source ) ) then
        return
    end if

    do i = 1, size(source)


        call setMod8Flag(source(i), image, qualityImage)

    end do


end subroutine SetMod8Flags


!******************************************************************
!
! Looks for a mod-8 pattern around a source
!
!******************************************************************
subroutine setMod8Flag(source, image, qualityImage)

    type(sourceParameters), intent(inout)    :: source
    real(single), pointer, dimension(:, :)   :: image
    integer(int16), pointer, dimension(:, :) :: qualityImage

    !*****************************************************************
    !
    ! Local variables
    !
    !*****************************************************************  
    integer(int16) :: imin, imax, jmin, jmax
    integer(int16) :: i, j, i1, i2, j1, j2, nx, ny, i0, j0, il, iu, jl, ju
    integer(int16) :: il1, iu1, jl1, ju1, n, nmin
    integer(int16) :: maxWidth
    integer(int16) :: iPlus1, jPlus1, iMinus1, jMinus1
    integer(int16) :: iPlus2, jPlus2, iMinus2, jMinus2
    integer(int16) :: iPlus3, jPlus3, iMinus3, jMinus3
    integer(int16) :: dx, n1, n3
    integer(int16) :: k1, k2, k3, k4, k5, kmin 
    real(single) :: fmax, fMin
    real(single), parameter :: nsigma = 2.0 ! was 2
    logical :: iSmod8Source


    !write(errstr, '(a, i6, a, f8.2, a, f8.2)') &
    !        "Checking for mod-8 pattern around source ", source%number, &
    !        " x=", source%xpos, " y=", source%ypos
    !call message(AppMsg, Verbose)
  



    isMod8Source = .false.    

    nx = ubound(image, 1)
    ny = ubound(image, 2)

    i0 = int(source%xpos - 0.5)
    j0 = int(source%ypos - 0.5)
    

    dx = 4 / binx  ! Was 16
    il = max( 0_int16, i0 - dx )
    iu = min( nx, i0 + dx )
    jl = max(0_int16, j0 - dx)
    ju = min(ny, j0 + dx)

    fMax = 0.0
    do i = il, iu
        do j = jl, ju
            if( .not. isNullValue(image(i, j)) ) then

                fmax = max( fMax, image(i, j) )
            end if
        end do
    end do 

    fmin = 0.1 * fmax 
   
    source%rawCountRate = -fmax

    dx = 30 / binx

    maxwidth = 16 / binx

    if (fmax .lt. 1000.0 ) then
        isMod8Source = .false.      
        return
    elseif(fmax .ge. 10000.0) then ! WAS 10000
        isMod8Source = .true. 
        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag

    else     
        isMod8Source = .false.
    end if
    

    imin = max(0_int16, i0 - maxWidth)
    imax = min(nx, i0 + maxWidth)
       
    jmin = max(0_int16, j0 - maxWidth)
    jmax = min(ny, j0 + maxWidth)

    il1 = max( 0_int16, i0 - dx )
    iu1 = min( nx,  i0 + dx )
    jl1 = max( 0_int16, j0 - dx )
    ju1 = min( ny,  j0 + dx )
  
    dx = 10 / binx ! was 6 / binx  

    il = max( 0_int16, i0 - dx )
    iu = min( nx,  i0 + dx )
    jl = max( 0_int16, j0 - dx )
    ju = min( ny,  j0 + dx )

    !****************************************************************
    !
    ! Check both sides of the source
    !
    !****************************************************************
    k1 = 0
    k2 = 0
    k3 = 0
    k4 = 0
  
    nmin = 2 * dx 

    !****************************************************************
    !
    ! Checking left-hand side of the source for a mod-8 pattern
    !
    !***************************************************************
    n = ju - jl + 1
   
    if (n .ge. 10) then
        nmin = n - 3
         
 iLoop1  : do i1 = i0 - 10 / binx, il1, -1
               
    
            iMinus1 = i1 - 1
            iMinus2 = i1 - 2
            iminus3 = i1 - 3
            if (iminus3 .lt. 0) exit               

            if ( any(image( i1 : i1, jl : ju) > fmax) ) exit

            !***********************************************************************************
            !
            ! Check that the column of pixels centred on iminus1 is less than the column
            ! centred on i1
            !
            !**********************************************************************************  
   
            n1 = count( image( i1 : i1, jl : ju) .gt. ( image( iminus1 : iminus1, jl : ju) + &
                 max(2.0, nsigma * sqrt( image( iminus1 : iminus1, jl : ju) ))) )
       

            if(n1 < nmin) cycle

            !*********************************************************************
            !
            ! Check that the column of pixels to the left of iminus1 is greater than
            ! the column centred on iminus1 
            !
            !********************************************************************            
            n3 = max( count( image(iMinus2 : iMinus2, jl : ju) .gt. image( iminus1 : iminus1, jl : ju) ), &
                      count( image(iMinus3 : iMinus3, jl : ju) .gt. image( iminus1 : iminus1, jl : ju) ))
       
            if ( n3 .ge. nmin) then

                !***********************************************************************
                !
                ! Maybe a mod-8 pattern to the left of the source
                !   
                !***********************************************************************           
                
                imin = min(imin, iminus3)
                k1 = 1

            end if
           
        end do iLoop1

     
        !****************************************************************
        !
        ! Checking right-hand side of the source for a mod-8 pattern
        !
        !***************************************************************    
 
iLoop2 : do i2 = i0 + 10 / binx, iu1   
        
            iPlus1 = i2 + 1
            iPlus2 = i2 + 2
            iplus3 = i2 + 3

            if (iplus3 .gt. nx) exit       


            if ( any(image( i2 : i2, jl : ju) > fmax) ) exit
     
            !*****************************************************************************
            !
            ! Check that the column of pixels centred on column i2 is greater than that
            ! centered on column iplus1
            !
            !*****************************************************************************
                    
            n1 = count( image(i2 : i2, jl : ju) .gt. ( image( iplus1 : iplus1, jl : ju) + &
                        max(2.0, nsigma * sqrt( image( iplus1 : iplus1, jl : ju) ))) )
           
            if(n1 .lt. nmin) cycle

            !*********************************************************************
            !
            ! Check that the column of pixels to the right of iplus11 is greater than
            ! the column centred on iplus11
            ! 
            !********************************************************************
            n3 = max( count( image(iplus2 : iplus2, jl : ju) .gt. image( iplus1 : iplus1, jl : ju)), &
                      count( image(iplus3 : iplus3, jl : ju) .gt. image( iplus1 : iplus1, jl : ju)) )
                
            if ( n3 .ge. nmin ) then


                !***********************************************************************
                !
                ! Maybe a mod-8 pattern to the right of the source
                !
                !***********************************************************************            
                imax = max(imax, iplus3)
                k2 = 1 

           
           end if

       
        end do iLoop2



    end if

    !****************************************************************
    !
    ! Check below and above the source
    !
    !****************************************************************
    

    n = iu - il + 1

    
    if (n .ge. 10) then
        nmin = n - 3
       
jLoop1 : do j1 = j0 + 10 / binx, ju1

            !****************************************************************
            !
            ! Checking above the source for a mod-8 pattern
            !
            !***************************************************************
    
            k5 = k5 + 1 
            jPlus1 = j1 + 1
            jPlus2 = j1 + 2
            jPlus3 = j1 + 3
       
            if (jplus3 .gt. ny) exit      
                                    
            if ( any(image( il : iu, j1 : j1) > fmax) ) exit
     
            !*****************************************************************************
            !
            ! Check that the row of pixels centred on row j1 is greater than that
            ! centered on column jplus1
            !
            !*****************************************************************************
            n1 = count( image(il : iu, j1 : j1) .gt. ( image( il : iu, jplus1 : jplus1) + &
                 max(2.0, nsigma * sqrt( image( il : iu, jplus1 : jplus1) ))) )
           
            if(n1 < nmin) cycle
          
            n3 = max( count( image(il : iu, jPlus2 : jPlus2) .gt. image( il : iu, jplus1 : jplus1)), &
                      count( image(il : iu, jPlus3 : jPlus3) .gt. image( il : iu, jplus1 : jplus1)) )
                     
            if (n3 .ge. nmin) then



                !***********************************************************************
                !
                ! Maybe a mod-8 pattern above the source
                !
                !***********************************************************************           
                
                jmax = max(jmax, jplus3)

                k3 = 1
            end if   
        end do jLoop1
   
   

        !****************************************************************
        !
        ! Checking below the source for a mod-8 pattern
        !
        !***************************************************************  
jLoop2 : do j2 = j0 - 10 / binx, jl1, -1
         
            jMinus1 = j2 - 1
            jMinus2 = j2 - 2
            jMinus3 = j2 - 3


            if (jMinus3 .lt. 0) exit     
           
 
         
            if ( any(image( il : iu, j2 : j2) > fmax) ) exit


            !*****************************************************************************
            !
            ! Check that the row of pixels centred on row j2 is greater than that
            ! centered on column jminus1
            !
            !****************************************************************************
            n1 = count( image(il : iu, j2 : j2) .gt. ( image( il : iu, jminus1 : jminus1) + &
                 max(2.0, nsigma * sqrt( image( il : iu, jminus1 : jminus1) ))) )

            if(n1 < nmin) cycle


            !*****************************************************************************
            ! Check that the row of pixels centred on row j1 is greater than that
            ! centered on column jplus1
            !*****************************************************************************
            n3 = max( count( image(il : iu, jMinus2 : jMinus2) .gt. image( il : iu, jminus1 : jminus1) ), &
                      count( image(il : iu, jMinus3 : jMinus3) .gt. image( il : iu, jminus1 : jminus1) ) )

            if ( n3 .ge. nmin) then

                !***********************************************************************
                !
                ! Maybe a mod-8 pattern below the source
                !
                !***********************************************************************           
                jmin = min(jmin, jminus3)
                            
            end if
                      
        end do jLoop2
    end if


    kmin = 3
    if (source%xpos <= real(12 / binx) .or. source%ypos <= real(12 / binx)) kmin = kmin  - 1
    if (source%xpos >= real(nx  - 12 / binx) .or. source%ypos >= real(ny  - 12 / binx)) kmin = kmin - 1
              

    if ( (k1 + k2 + k3 + k4) .ge. kmin ) then ! Accept it is as having a mod-8 pattern 

        IsMod8Source = .true.
           
    end if


    if(isMod8Source) then
        !********************************************************************************
        !
        ! 
        !********************************************************************************

        source%qualityFlag = ibset(source%qualityFlag, 4) ! Set bit 4 of quality flag



 !       write(*, *) "setmod8flag imin=", imin, " imax=", imax, " jmin=", jmin, " jmax=", jmax

        do i = iMin, iMax
            do j = jMin, jMax  

                qualityImage(i, j) = ibset(qualityImage(i, j), 4) ! Set bit 4 of quality flag
            end do
        end do


    end if


end subroutine SetMod8Flag





!***************************************************
!
!  subroutine 
!
!***************************************************
Subroutine SetBit8Flags( nSources, source, qualityImage)
  
    integer, intent(in) :: nSources
    type(sourceParameters), pointer :: source(:)
  
    integer(int16), pointer :: qualityImage(:, :)

    integer :: i
    if( .not. associated( source ) ) then
        return
    end if

    do i = 1, size(source)


        call setQualityImageBit8Pixels(source(i), qualityImage)

    end do

end subroutine SetBit8Flags



!******************************************************************
!
! BIT-8 Point-source within an extended source
!
!******************************************************************
subroutine setQualityImageBit8Pixels(source, qualityImage)

    type(sourceParameters), intent(inout)    :: source
    integer(int16), pointer, dimension(:, :) :: qualityImage

    !*****************************************************************
    !
    ! Local variables
    !
    !*****************************************************************
    integer(int16) :: i, j, imin, imax, jmin, jmax, nx, ny
    
    real(single) :: semiA, semiB, theta, x, y
    real(single), parameter :: nstd = 2.0
    if( source%extensionFlag .ne. 1 ) then
        return 
    end if


  write(*, *) "addflags_mod.f90 line 1129 - setting bit 8 flags on image"



    nx = ubound(qualityImage, 1)
    ny = ubound(qualityImage, 2)


    semiA = nstd * source%semia
    semiB = nstd * source%semib

    if(source%theta < 0.0) then
        theta = source%theta + 180.0
    else
        theta = source%theta
    end if
    theta = theta * degreesToRadians
    imin = max(0, floor(source%xpos - semiA) )
    imax = min(nx, int(floor(source%xpos + semiA), int16) )
    jmin = max(0, floor(source%ypos - semiA) )
    jmax = min(ny, int(floor(source%ypos + semiA), int16) )

  
    do i = imin, imax
        x = real(i + 1)
        do j = jmin, jmax
            y = real(j + 1)
            if (PointInEllipse(source%xpos, source%ypos, x, y, &
                   semiA, semiB, theta) ) then
                qualityImage(i, j) = ibset(qualityImage(i, j), 8)                 



            end if


        end do
    end do    

end subroutine setQualityImageBit8Pixels

!***************************************************
!
!  subroutine SetBit9Flags
!  Looks for isolated very bright pixels
!
!***************************************************
Subroutine SetBit9Flags( image, qualityImage, halfWidth)
  
    
    real(single), pointer :: image(:, :)
    integer(int16), pointer :: qualityImage(:, :)
    integer(int8), intent(in) :: halfWidth 
    integer(int16) :: i, j, k, l, i1, j1, nx, ny, il, iu, jl, ju, count
    real(single) :: f1, f2, f3
   
    nx = ubound(image, 1)
    ny = ubound(image, 2)

 
    write(errstr, '(a, i6)') "setting bit 9 on quality-image pixels"
    call message(AppMsg, Verbose)
  
    do i = 0, nx
        do j = 0, ny
            f1 = image(i, j) 
      
            if ( isNullValue(f1)  ) cycle
            if( f1 < 100.0) cycle 
            count  = 0
            f3 = f1 * 0.05
            do k = max(0_int16, i - 1_int16), min(nx, i + 1_int16)
                do l = max(0_int16, j - 1_int16), min(ny, j + 1_int16)
                    f2 = image(k, l) 
      
                    if ( isNullValue(f2)  ) cycle
                    if(f2 >= f3) count = count + 1 
                end do
            end do 
         
            if( count < 2) then

                !*************************************************************
                !
                ! Set pixel bits within box
                !
                !*************************************************************
                il = max(0_int16, i - halfWidth)
                jl = max(0_int16, j - halfWidth)
                iu = min(nx, i + halfWidth)
                ju = min(ny, j + halfWidth)

                do i1 = il, iu
                    do j1 = jl, ju 
 
                        qualityImage(i1, j1) = ibset(qualityImage(i1, j1), 9)                 
                    end do
                end do
            end if
        end do
    end do

end subroutine SetBit9Flags

end module addflags_mod
