/************************************************************************************
*
* Contains function for setting quality-flag bits on a 16-bit QUALITY image
* using both the PRIMARY image and the source-list created by omdetect
*
*
*************************************************************************************/

#include "omqualitymapdefs.h"  

#include <iostream>
#include <vector>       
#include "sourcelist.h"  // For input/output to/from FITS files
#include "utility.h"     // 

/***********************************************************************
*
*
*  Sets bit 5 of all the pixels within the central enhancement region
*
*
**********************************************************************/
void setCentralRegionPixels( const int x0, const int y0, const int binX, vector<vector<int16> > & image)
{





    errstr << "Setting bit 5 of the quality-image pixels" << Sas::message(Msg::App, Msg::Verbose);
 

    const float xc = 992.0, yc = 1063.0, radius = 160.0;   
    const int nx = (int)image.size();
    const int ny = (int)image[0].size();

    for(int j = 0; j < ny; j++)
    {
      
        float y = float(j + 1); 
        for(int i = 0; i < nx; i++)
	{
	    if( image[i][j] < 0) continue;

	    float x = float(i + 1);
            float x1, y1; 
            ConvertCoordsToFullFrame(x, y, x0, y0, binX, binX, x1, y1);
            double dx = x1 - xc, dy = y1 - yc;      
            if (hypot(dx, dy) < radius)
	    {
	                   
	        setBit( image[i][j], 5);  // Set bit 5	            
                       
            }
	}

    }

}

/***********************************************************************
*
* Sets bit 7 of all the pixels near to an image edge
*
*
*
**********************************************************************/
void setEdgePixels( const int binX, vector<vector<int16> > & image)
{

    errstr << "Setting bit 7 of the quality-image pixels" << Sas::message(Msg::App, Msg::Verbose);
 
    const int nx = (int)image.size();
    const int ny = (int)image[0].size();


    const int width = 12 / binX;
    for(int i = 0; i < width; i++)
    {
 
        for(int j = 0; j < ny; j++)
            setBit( image[i][j], 7);  // Set bit 7	            
     

    }

    for(int i = nx - width; i < nx; i++)
    {
 
        for(int j = 0; j < ny; j++)
            setBit( image[i][j], 7);  // Set bit 7	            
     
 
    }

    for(int i = 0; i < width; i++)
    {
 
        for(int j = 0; j < nx; j++)
            setBit( image[j][i], 7);  // Set bit 7	            
     
    }

    for(int i = ny - width; i < ny; i++)
    {
 
        for(int j = 0; j < nx; j++)
            setBit( image[j][i], 7);  // Set bit 7	                 
 
    }

}

/**************************************************************************************
*
* Sets bit 7 of all pixels near the extreme corners of the OM detector window
*
****************************************************************************************/
void setCornerPixels( const int x0, const int y0, const int binX, vector<vector<int16> > & image)
{

    errstr << "Setting bit 5 of the quality-image pixels" << Sas::message(Msg::App, Msg::Verbose);
 

    const float xc = 979.2, yc = 1016.0, radius = 1300.0;
   

    int nx = (int)image.size();
    int ny = (int)image[0].size();

    for(int j = 0; j < ny; j++)
    {
      
        float y = float(j + 1); 
        for(int i = 0; i < nx; i++)
	{
	    if( image[i][j] < 0) continue;
	    float x = float(i + 1);
            float x1, y1; 
            ConvertCoordsToFullFrame(x, y, x0, y0, binX, binX, x1, y1);
            double dx = x1 - xc, dy = y1 - yc;      
            if (hypot(dx, dy) > radius)
	    {               
	        setBit( image[i][j], 7);  // Set bit 7	            
                       
            }
	}

    }

 

}






